# CSSMinMaxR24.py for Notepad++ Python Scripting plugin * For CSS developers and web designers
# This is a Python port of the YUI CSS Compressor so you can minify/format CSS (JS not tested)
# Based on CSSmin.py v.1.0 from https://github.com/ethanpil/npp-cssmin
#
## Extends 2024
# Patched by Barmaley: bugs fixed + useful features added * GPLv3 * GoodCRX Project
#
# 2024.8.31 (initial fork):
# 1. No required space for @media ") or (" = fixed.
# 2. No required spaces in calc with "+" sign 'calc(1px + 1em)' = fixed.
# 3. Preserved comments now defined with /**..*/ (instead of original /*!..*/)
# 4. If some text selected in active file: process fragment minification in current document (as in original script, see Usage below)
# 5. If nothing selected in active file: force process minification in NEW document (create new than process the copy of source CSS)
# 6. After processing autosave* new compressed file, autonamed from original name with change to "_.css"
#	 * Note: if any file with same name exist in the same dir - it's overwrite without confirmation
# 7. Force pre-save current focument backup copy* (create in same dir as opened file, add .BAK to current filename)
#	 * This copy overwrite every time. No way in NPP to automatically remove temp backup files. Delete them manually if need.
#
# 2024.8.30 -> 2024.8.31:
# + Compressed CSS: all preserved comments delimits by new line from previous/following CSS rules (\n)
# + Fix logic: reset UNDO ops ONLY if minified "by place" (if process run in current file, else undo-buffer stay invalid)
# + Fix/improved: regexp for "calc(x + y)" expressions and for and/or conditions (e.g. @media/@import etc.)
# + Improve/optimize: multi-to-one zeros conversion, multi-semicolon remover, zero-value with units (px/vh/..) fixer
#
# 2024.8.31 -> 2024.9.4:
# + Rebrand project: nows it's named as CSSMinMaxR24
# + Simplest CSS autoformatter: autodetect compressed CSS then expand it - every rules converted to indented CSS blocks:
# 	* If some text selected: replace formatted fragment "by place" (as compressed in previous versions)
# 	* If nothing selected: current compessed CSS expands, create new autonamed file with *_MAX.css as name
#	* Yet works fine only with usual and one-level conditional rules (e.g. @media), otherwise text indents may be not "beauty"
# + Fix/improved: regexp for "calc(x + y)" expressions and for and/or conditions (e.g. @media/@import etc.)
# + Fix/improved: check/autofix for missing/excess :: in pseudo-elements (::after/etc) + fix excess :: in pseudo-classes (:hover/etc)
# + Improve/optimize: multi-to-one zeros conversion, multi-semicolon remover, zero-value with units (px/vh/..) fixer
# + Simplest "CSS-beatifier": add some spaces after/before delimiters (e.g. ,_ :_ _{ ), no trailing spaces and empty lines
# + Source code optimization: some "one-time-call" function moved to aproriate execution points
#
## Installation
#	Simply copy "CSSMinMax.py" to "..\plugins\PythonScript\scripts\" in your NPP folder and restart NPP.
#	It will now appear as a menu item under Plugins > Python Script > Scripts
#
## Usage
#	1. Select text to process, or select nothing.
#	2. Go to the NPP menu, Plugins...Python Script...Scripts...CSSmin and click!
#	3. If selected text detected, it will process and replace the selected text, otherwise the entire contents of current document.
#	4. Undo is available if you dont like the results
#	5. Also you can check the output CSS for errors online, e.g. here https://jigsaw.w3.org/css-validator/
#
##Credits
#	Original cssmin.py ported from YUI here https://github.com/zacharyvoase/cssmin
#
##Licence
#	Original BSD changed to GPLv3 for this fork and it's next childs.
###

# At first, create source file backup... :)
notepad.saveAsCopy((notepad.getCurrentFilename() + ".BAK"))

from StringIO import StringIO # The pure-Python StringIO supports unicode.
import re

tmp = ["_","","","","","",""]	# Array of temporary strings (used/changed/erased in both minifier/expander)

# Remove CSS comments (except preserved).
def remove_comments(css):
	iemac = False
	preserve = False
	comment_start = css.find("/*")
	# Slicing to avoid an IndexError.
	while comment_start >= 0:
		# Preserve comments like /**...*/ (two stars ** followed open slash).
		preserve = css[comment_start + 2:comment_start + 3] == "*"
		comment_end = css.find("*/", comment_start + 2)
		if comment_end < 0:
			if not preserve:
				css = css[:comment_start]
				break
		elif comment_end >= (comment_start + 2):
			# Check for IE Mac-specific comment, leave this one and the following one alone.
			if css[comment_end - 1] == "\\":
				comment_start = comment_end + 2
				iemac = True
			elif iemac:
				comment_start = comment_end + 2
				iemac = False
			elif not preserve:
				css = css[:comment_start] + css[comment_end + 2:]
			else:
				comment_start = comment_end + 2
		comment_start = css.find("/*", comment_start)
	# Add preserved comments dumb: replaced to \n at last
	tmp[1] = "___PRESCOMM___"
	return re.sub(r"^"+tmp[1], "", re.sub(r"(.+)?\/\*\*", tmp[1]+"/**", re.sub(r"\*\/\s*", "*/"+tmp[1], css)))

# Remove excess whitespace characters.
def remove_excess_whitespace(css):
	# Prevents 'p :link' from becoming 'p:link': translates 'p :link' into 'p ___PSEUDOCLASSCOLON___link', this is translated back later.
	def pseudoclasscolon(css):
		tmp[2] = "___PSEUDOCLASSCOLON___"
		regex = re.compile(r"(^|\})(([^\{\:])+\:)+([^\{]*\{)")
		match = regex.search(css)
		while match:
			css = ''.join([
				css[:match.start()],
				match.group().replace(":", tmp[2]),
				css[match.end():]])
			match = regex.search(css)
		return css

	# Remove spaces from before things.
	css = pseudoclasscolon(css)
	css = re.sub(r"\s+([!{};:>+\(\)\],])", r"\1", css)

	# If there is a @charset, then only allow one, and move to the beginning.
	css = re.sub(r"^(.*)(@charset \"[^\"]*\";)", r"\2\1", css)
	css = re.sub(r"^(\s*@charset [^;]+;\s*)+", r"\1", css)

	# Put the space back: @media screen and/or (min-width:100px).
	css = re.sub(r"\b(and|or)\(", r"\1 (", css)
	# css = re.sub(r"\bor\(", "or (", css)

	# Put the colons back.
	css = css.replace(tmp[2], ':')

	# Remove spaces from after things.
	css = re.sub(r"([!{}:;>+\(\[,])\s+", r"\1", css)

	# Put the space back: calc(100vw + 100%).
	css = re.sub(r"\b(\d+("+tmp[5]+"))\+", r"\1 + ", css)
	return re.sub(r"\b\)\+", ") + ", css)

# Convert rgb(51,102,153) to #336699.
def normalize_rgb_colors_to_hex(css):
	regex = re.compile(r"rgb\s*\(\s*([0-9,\s]+)\s*\)")
	match = regex.search(css)
	while match:
		colors = map(lambda s: s.strip(), match.group(1).split(","))
		hexcolor = '#%.2x%.2x%.2x' % tuple(map(int, colors))
		css = css.replace(match.group(), hexcolor)
		match = regex.search(css)
	return css

# Replace :0 0 0 0;, :0 0 0; etc. with :0;.
def condense_multidimensional_zeros(css):
	css = re.sub(r"\b:(\s*)?(0\s+){1,3}0;", ":0;", css)
	# Revert background-position:0; to the valid value.
	css = css.replace("background-position:0;", "background-position:0 0;")
	return css

# Shorten colors from #AABBCC to #ABC where possible.
def condense_hex_colors(css):
	regex = re.compile(r"([^\"'=\s])(\s*)#([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])([0-9a-fA-F])")
	match = regex.search(css)
	while match:
		first = match.group(3) + match.group(5) + match.group(7)
		second = match.group(4) + match.group(6) + match.group(8)
		if first.lower() == second.lower():
			css = css.replace(match.group(), match.group(1) + match.group(2) + '#' + first)
			match = regex.search(css, match.end() - 3)
		else:
			match = regex.search(css, match.end())
	return css

# Wrap the lines of the given CSS to an approximate length.
def wrap_css_lines(css, line_length):
	lines = []
	line_start = 0
	for i, char in enumerate(css):
	# It's safe to break after } characters.
		if char == '}' and (i - line_start >= line_length):
			lines.append(css[line_start:i + 1])
			line_start = i + 1
	if line_start < len(css):
		lines.append(css[line_start:])
	return '\n'.join(lines)

# Automatic fix for missing/excess ::.
def pseudo_elements_sign(css):
	tmp[4] = "(after|before|marker|selection|first-letter|first-line|grammar-error|spelling-error)"
	css = re.sub(r"\s?[^:]:("+tmp[4]+"[^\w-])", r"::\1", css)
	css = re.sub(r"(:\s*){2,}("+tmp[4]+"[\w-])", r":\2", css)
	return re.sub(r"(\s*:){2,}\s*", "::", css)

# CSS-compressor sequence.
def cssmin(css, wrap=None):
	css = remove_comments(css)
	# Condense multiple whitespaces into one.
	css = re.sub(r"\s+", " ", css)
	# A pseudo class for the Box Model Hack (http://tantek.com/CSS/Examples/boxmodelhack.html)
	tmp[3] = "___PSEUDOCLASSBMH___"
	css = css.replace('"\\"}\\""', tmp[3])
	tmp[5] = "%|px|vh|vw|vmin|vmax|em|rem|cm|mm|in|pt|pc|ex|ch"
	css = remove_excess_whitespace(css)
	# Replace 0(px, em, %, etc) with 0.
	css = re.sub(r"([\s:])(0)("+tmp[5]+"|)", r"\1\2", css)
	css = condense_multidimensional_zeros(css)
	# Replace 0.6 with .6 where possible.
	css =  re.sub(r"(:|\s)0+\.(\d+)", r"\1.\2", css)
	css = normalize_rgb_colors_to_hex(css)
	css = condense_hex_colors(css)
	if wrap is not None:
		css = wrap_css_lines(css, wrap)
	css = css.replace(tmp[3], '"\\"}\\""')
	# Condense multiple ";;;" into one and remove trailing.
	css = re.sub(r";+\}", "}", re.sub(r";+", ";", css))
	# Return back the preserved comments with new line indent.
	css = re.sub(tmp[1], r"\n", css)
	return css.strip()

# Basic CSS formatter, parse/indents "usual" rules
def base_indent(css):
	tmp[1] = "___AT_COND___"
	tmp[2] = "___AT_RULES_SET___"
	tmp[3] = "___AT_RULES_END___"
	css = re.sub(r'(\*\/)', r'\1\n', css)	# Comments with new lines
	css = re.sub('@', tmp[1], css)	# Preset: conditional rules declare
	css = re.sub(r'('+tmp[1]+'[^\{]+)\{', r'\1'+tmp[2]+'\n', css)	# Preset: conditional rules begin
	css = re.sub('}}', '}'+tmp[3]+'\n', css)	# Preset: conditional rules end
	css = re.sub(r'([^({})]+\{)', r'\1\n\t', css)	# Rules begin
	css = css_beauty(css, 1)	# Add one spaces after ":"/"," in declares when css-format is basic only
	css = re.sub(r';', r';\n\t', css)		# Next declare
	css = re.sub(r'(?!"\'[^;]*);(?![^;]*\'")', r';\n\t', css)		# Next declare
		# css = re.sub(r'(\t*\})', r'\n\1\n', css)	# Rules end
	css = re.sub(r'(\t*\})', r'\n}\n', css)	# Rules end
	return css_beauty(css, 2)	# Add one space after ":" in declares and absent ";" at last declaration end

# Conditional rules formatter (parse/indents @media/etc. attached rules blocks)
def cond_indent(css):
	# Note: shift yet works fine ONLY for ONE-level conditional rules!
	def rule_shift(css, arr, cond, shift):
		for rule in arr:
			# Check for rules block is inside the condition
			if re.search(cond, rule):
				# Add more one tab-shift to all internal rules and declares strings
				css = css.replace(rule, rule.replace('\n', '\n'+('\t'*shift)))
		return css

	rgx = r'('+tmp[3]+'|'+tmp[2]+')'
	css = rule_shift(css, re.split(tmp[1], css), tmp[3], 1)	# Note: now shift works fine ONLY for ONE-level conditional rules!
	# Revert back temp conditional rules presets to source
	css = re.sub(r'\t*'+tmp[1], '@', css)	# Revert declares preset
	css = re.sub(tmp[2], ' {', css)			# Revert begins preset
	css = re.sub(r'\t*'+tmp[3], '}', css)		# Revert ends preset
	return css

# CSS beautifier: increase readability of expanded text.
def css_beauty(css, step):
	if step == 1:	# Pre-improve when css-format is basic only (avoid unwanted issues at next process steps)
		css = re.sub(r'(["\']data:)(.+;)(base64,)', r'\1 \2\n\t\t\3 ', css)	# Experimental: add space after ":"/"," in base64-urls
	if step == 2:	# Pre-improve before run conditional formatter, after all basic formats applied (avoid issues at next steps)
		css = re.sub(r'(\n\t[^:]+:)', r'\1 ', css)	# Add absent space after ":" in declares
		css = re.sub(r'([^\};])\n\}', r'\1;\n}', css)	# Add absent ";" at last declaration end
	else:	# Final improvments before 
		css = re.sub('{', ' {', css)	# Add spaces before "{" for best readability
		css = re.sub(r'(?!\'data:\s?image\/[a-z]{3,4};base64),', ', ', css)	# Add spaces after "," for best readability
		# css = re.sub(r'(\n(\t*)?)+\n', r'\n', css)	# Remove exceed lines
		css = re.sub(r'(\n\s*)+\n', r'\n', css)	# Remove exceed empty lines
		css = re.sub(r'(\s+)$', '', re.sub(r'[ ]{2,}', ' ', css))	# Remove exceed spaces and trailing spaces/tabs
		css += "\n"	# Add line at the EOF
	return css

# CSS expander/formatter: expand compressed CSS to human-view format.
def cssmax(css):
	css = base_indent(css)		# Basic formatter, "usual" rules
	css = cond_indent(css)		# Conditional rules formatter
	css = css_beauty(css,"")	# Improve readability
	return css

# Preprocess by flag max|min + fix pseudo-elements rules.
def outcss(css):
	# Set process direction flag - max|min.
	max = re.search("\}((\n[^\/])|([^\/]\n))", css)
	if max:	# Set global temp-strings array for CSS-minifier
		css = cssmin(css)
	else:	# Set global temp-strings array for CSS-expander
		tmp[0] = "_MAX"
		css = cssmax(css)
	return pseudo_elements_sign(css)

## All functions defined. Now execute.
notepad.save()	# Force autosave source file, anyway.

# Is selected text - process that, otherwise - process entire document
csstxt = editor.getSelText()

if len(csstxt) >= 1 :	# If selection in current document: process fragment "by place"
	editor.beginUndoAction()	# Undo ONLY if by place - preserv current document undo buffer!
	editor.replaceSel(outcss(csstxt))
	editor.endUndoAction()	# Undo ONLY if by place - preserv current document undo buffer!
else:	# If nothing selected: set the new filename to create copy as new document
	tmp[5] = notepad.getCurrentFilename()
	tmp[6] = outcss(editor.getText())
	notepad.new()
	editor.setText(tmp[6])
	notepad.saveAs(re.sub(r"\.css$", tmp[0]+".css", tmp[5]))

tmp = ["_","","","","","",""]	# Reset/cleanup temp-strings array

# That's all, folks! :)